import numpy as np
from enum import auto
from typing import Tuple

from systemstoolkit.files.keywords import KeywordEnum
from systemstoolkit.files.validators import (
    DataValidator,
    QuaternionValidator,
    AngleValidator,
    NoValidator,
)


class DataFileFormat(KeywordEnum):
    def validate_data(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return self.validator.validate(time, data)


class AttitudeFileFormat(DataFileFormat):
    # Quaternion-based formats
    Quaternions = auto()
    QuatScalarFirst = auto()
    QuatAngVels = auto()
    AngVels = auto()

    # Euler-based Formats
    EulerAngles = auto()
    EulerAngleRates = auto()
    EulerAnglesAndRates = auto()

    # YPR-based formats
    YPRAngles = auto()
    YPRAngleRates = auto()
    YPRAnglesAndRates = auto()

    # DCM-based formats
    DCM = auto()
    DCMAngVels = auto()

    # Vector-based formats
    ECFVector = auto()
    ECIVector = auto()

    def __str__(self) -> str:
        return f'AttitudeTime{self.name}'

    @property
    def validator(self) -> DataValidator:
        if self.name in ['Quaternions', 'QuatScalarFirst']:
            return QuaternionValidator()
        elif self.name in ['EulerAngles', 'YPRAngles']:
            return AngleValidator()
        else:
            return NoValidator()



import datetime
import numpy as np
from enum import Enum, auto
from typing import Union
from abc import ABC

from systemstoolkit.typing import Union, DateTimeLike, DateTimeArrayLike


class TimeUnit(ABC):
    def convert(
            self,
            time: Union[DateTimeLike, DateTimeArrayLike]
        ) -> Union[DateTimeLike, DateTimeArrayLike]: # pragma: no cover
        pass


class EpochTimeUnit(TimeUnit):
    unit = 's'

    def __init__(
            self, 
            epoch: datetime.datetime
        ) -> None:
        super().__init__()
        self.epoch = epoch

    @property
    def epoch(self):
        return self._epoch

    @epoch.setter
    def epoch(self, epoch: DateTimeLike):
        self._epoch = np.datetime64(epoch)

    def convert(
            self,
            time: Union[DateTimeLike, DateTimeArrayLike],
        ):
        time = np.asarray(time, dtype='datetime64[ms]')
        return (time - self.epoch) / np.timedelta64(1, self.unit)


class EpSecTimeUnit(EpochTimeUnit):
    unit = 's'


class EpMinTimeUnit(EpochTimeUnit):
    unit = 'm'


class EpHrTimeUnit(EpochTimeUnit):
    unit = 'h'


class EpDayTimeUnit(EpochTimeUnit):
    unit = 'D'


# class EpYearTimeUnit(EpochTimeUnit):
#     unit = 'Y'


class UTCTime(TimeUnit):
    pass


class YYYYDDDTimeUnit(UTCTime):
    def convert(
            self,
            time: Union[DateTimeLike, DateTimeArrayLike]
        ) -> Union[DateTimeLike, DateTimeArrayLike]:

        time = np.asarray(time, dtype='datetime64[ms]')

        day = time.astype('datetime64[D]')
        year = day.astype('datetime64[Y]')
        doy = day - year
        frac_day = (time - day) / np.timedelta64(1, 'D')

        yyyy = 1000*(1970+year.astype('uint32'))
        ddd = doy.astype('uint32') + 1 + frac_day.astype('float64')
        return yyyy + ddd


class YYYYMMDDTimeUnit(UTCTime):
    def convert(
            self,
            time: Union[DateTimeLike, DateTimeArrayLike]
        ) -> Union[DateTimeLike, DateTimeArrayLike]:

        time = np.asarray(time, dtype='datetime64[us]')
        day = time.astype('datetime64[D]')
        mon = time.astype('datetime64[M]')
        year = time.astype('datetime64[Y]')
        moy = mon - year
        dom = day - mon
        frac_day = ((time - day) / np.timedelta64(1, 'D')).astype('float64')
        y = 10000 * (year.astype('uint32') + 1970)
        m = 100 * (moy.astype('uint32') + 1)
        d = dom.astype('uint32') + 1
        print(y, m, d, frac_day)
        return y + m + d + frac_day


class ISOYMDTimeUnit(UTCTime):
    def convert(
            self,
            time: Union[DateTimeLike, DateTimeArrayLike]
        ) -> Union[DateTimeLike, DateTimeArrayLike]:

        time = np.asarray(time, dtype='datetime64[ms]')
        return time.astype(str)



import io
import numpy as np
from typing import Union, Optional
from dataclasses import dataclass, asdict
from enforce_typing import enforce_types

from .formats import AttitudeFileFormat
from .keywords import (
    Keyword,
    AttitudeDeviations,
    ScenarioEpoch,
    Coordinate,
    CoordinateAxes,
    CoordinateAxesEpoch,
    MessageLevel,
    CentralBody,
    BlockingFactor,
    Interpolation,
    InitialAttitude,
    TimeFormat,
    TrendingControl,
    NumberOfAttitudePoints,
)


class StkFile:
    version = '11.0'


ATTITUDE_FILE_TEMPLATE = '''stk.v.{version}
BEGIN Attitude
{keywords}
{format}
{data}
END Attitude
'''


@enforce_types
@dataclass
class AttitudeFile(StkFile):
    time: Union[list, np.ndarray]
    data: Union[list, np.ndarray]
    format: AttitudeFileFormat
    epoch: Optional[ScenarioEpoch] = None
    message: Optional[MessageLevel] = MessageLevel('Warnings')
    axes: Optional[Coordinate] = Coordinate(CoordinateAxes('ICRF'))
    body: Optional[CentralBody] = CentralBody('Earth')
    interp: Optional[Interpolation] = None
    deviations: Optional[AttitudeDeviations] = None
    blocking: Optional[BlockingFactor] = None
    initial: Optional[InitialAttitude] = None
    time_fmt: Optional[TimeFormat] = TimeFormat('EpSec')
    trending: Optional[TrendingControl] = None

    def __post_init__(self):
        self.time, self.data = self.format.validate_data(self.time, self.data)
        self.points = NumberOfAttitudePoints(self.data.shape[0])

    def keywords(self) -> str:
        lines = []
        for key in asdict(self).keys():
            keyword_obj = getattr(self, key)
            if isinstance(keyword_obj, Keyword):
                lines.append(str(keyword_obj))
        return '\n'.join(lines) + '\n'

    def format_data(self) -> str:
        buf = io.StringIO()

        if self.epoch is None:
            self.epoch = ScenarioEpoch(self.time[0])
        
        formatted_time = self.time_fmt.convert(self.time, epoch=self.epoch.value)
        for t, row in zip(formatted_time, self.data):
            print(
                str(t).rjust(15),
                ' '.join([f'{x:.6f}'.rjust(15) for x in row]),
                file=buf
            )
        
        return buf.getvalue()

    def to_string(self) -> str:
        # Format data first to set epoch if necessary
        formatted_data = self.format_data()

        return ATTITUDE_FILE_TEMPLATE.format(
            version = self.version,
            keywords = self.keywords(),
            format = self.format,
            data = formatted_data,
        )



import numpy as np
from enum import Enum, auto
from dataclasses import dataclass, asdict
from typing import Optional, Union

from systemstoolkit.utils import stk_datetime
from systemstoolkit.typing import DateTimeLike, ArrayLike
from systemstoolkit.units.time import (
    TimeUnit,
    EpSecTimeUnit,
    EpMinTimeUnit,
    EpHrTimeUnit,
    EpDayTimeUnit,
    # EpYearTimeUnit,
    YYYYDDDTimeUnit,
    YYYYMMDDTimeUnit,
    ISOYMDTimeUnit,
)


KEYWORD_WIDTH = 30


class Keyword:
    @property
    def keyword(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        return f'{str(self.keyword).ljust(KEYWORD_WIDTH)} {self.value}'


class CompoundKeyword(Keyword):
    def __str__(self) -> str:
        # Iterate through the dataclass fields
        # returning formatted keywords that are not None
        keywords = [getattr(self, key) for key in asdict(self)]
        return '\n'.join([str(key) for key in keywords if key is not None])


class KeywordEnum(Enum):
    @classmethod
    def _missing_(cls, name):
        for member in cls:
            if member.name.lower() == name.lower():
                return member
    
    def _generate_next_value_(name, start, count, last_values):
        return name

    def __str__(self) -> str: # pragma: no cover
        return f'{str(self.keyword).ljust(KEYWORD_WIDTH)} {self.name}'


class MessageLevel(Keyword, KeywordEnum):
    Errors = auto()
    Warnings = auto()
    Verbose = auto()


class CentralBody(Keyword, KeywordEnum):
    Earth = auto()
    Moon = auto()


class CoordinateAxes(Keyword, KeywordEnum):
    ICRF = auto()
    J2000 = auto()
    Inertial = auto()
    Fixed = auto()
    TrueOfDate = auto()
    MeanOfDate = auto()
    TEMEOfDate = auto()
    B1950 = auto()
    MeanOfEpoch = auto()
    TrueOfEpoch = auto()
    TEMEOfEpoch = auto()
    AlignmentAtEpoch = auto()


class AttitudeDeviations(Keyword, KeywordEnum):
    Rapid = auto()
    Mild = auto()


class InterpolationMethod(Keyword, KeywordEnum):
    Lagrange = auto()
    Hermite = auto()


@dataclass
class InterpolationOrder(Keyword):
    value: int = 1


@dataclass
class Epoch(Keyword):
    value: DateTimeLike

    def __post_init__(self):
        self.value = np.datetime64(self.value)

    def __str__(self) -> str:
        return f'{str(self.keyword).ljust(KEYWORD_WIDTH)} {stk_datetime(self.value)}'


class ScenarioEpoch(Epoch):
    pass


class CoordinateAxesEpoch(Epoch):
    pass


@dataclass
class Coordinate(CompoundKeyword):
    axes: CoordinateAxes
    epoch: Optional[CoordinateAxesEpoch] = None

    def __post_init__(self):
        if 'epoch' in self.axes.value.lower():
            if self.epoch is None:
                raise ValueError(f'CoordinateAxes "{self.axes.value}" requires a CoordinateAxesEpoch')
        else:
            if self.epoch is not None:
                raise ValueError(f'CoordinateAxes "{self.axes.value}" does not support CoordinateAxesEpoch')


@dataclass
class Interpolation(CompoundKeyword):
    method: InterpolationMethod
    order: Optional[InterpolationOrder]


@dataclass
class NumberOfAttitudePoints(Keyword):
    value: int = None


@dataclass
class BlockingFactor(Keyword):
    value: int = None


class InitialAttitude(Keyword):
    pass


# class KeywordObjectEnum(KeywordEnum):

class TimeFormat(Keyword, KeywordEnum):
    EpSec = EpSecTimeUnit
    EpMin = EpMinTimeUnit
    EpHour = EpHrTimeUnit
    EpDays = EpDayTimeUnit
    # EpYear = EpYearTimeUnit
    YYYYDDD = YYYYDDDTimeUnit
    YYYYMMDD = YYYYMMDDTimeUnit
    ISOYMD = ISOYMDTimeUnit

    @classmethod
    def _missing_(cls, name):
        for member in cls:
            if member.name.lower() == name.replace('-', '').lower():
                return member

    def convert(self, time: ArrayLike, epoch=None) -> ArrayLike:
        time_fmt = self.value(epoch)
        return time_fmt.convert(time)

    def __str__(self) -> str: # pragma: no cover
        return f'{str(self.keyword).ljust(KEYWORD_WIDTH)} {self.name}'


class TrendingControl(Keyword):
    pass



import datetime
import numpy as np

from .files import AttitudeFile
from .formats import AttitudeFileFormat
from .keywords import (
    MessageLevel,
    Coordinate,
    CoordinateAxes,
    CoordinateAxesEpoch,
    ScenarioEpoch,
    CentralBody,
    Interpolation,
    InterpolationMethod,
    InterpolationOrder,
    AttitudeDeviations,
    BlockingFactor,
)


def attitude_file(
    time: np.ndarray,
    data: np.ndarray,
    format: str = None,
    epoch: datetime.datetime = None,
    axes: str = None,
    axes_epoch: datetime.datetime = None,
    message: str = 'Warnings',
    body: str = None,
    int_method: str = None,
    int_order: int = None,
    deviations: str = None,
    blocking: int = None,
) -> str:
    '''Create an STK Attitude (.a) file.
    
    Params
    ------
    time: np.ndarray[np.datetime64]

    data: np.ndarray[][]

    format: str
        The file formats used to specify data points in the attitude file.
        (i.e. "AttitudeTimeQuaternions" keyword for Quaternions)

        Quaternion-based formats
            Quaternions
            QuatScalarFirst
            QuatAngVels
            AngVels
        Euler-based Formats
            EulerAngles
            EulerAngleRates
            EulerAnglesAndRates
        YPR-based formats
            YPRAngles
            YPRAngleRates
            YPRAnglesAndRates
        DCM-based formats
            DCM
            DCMAngVels
        Vector-based formats
            ECFVector
            ECIVector
    
    epoch: datetime.datetime
        ScenarioEpoch. This is the reference epoch time for the time values of the attitude data in the file. 
        
        There is no relationship between ScenarioEpoch and the scenario epoch in your STK scenario.
        
        The default is the time in time[0].
    
    axes: str
        CoordinateAxes. AGI recommends that you supply these, but they are not required.

        These are the reference axes from which the rotation to the body axes is performed. Normally, the coordinate axes is the name of a valid coordinate system for the central body specified above (see Central Body Coordinate Systems). Typically, each central body supports Fixed, J2000, ICRF, Inertial, TrueOfDate, and MeanOfDate, but some bodies support additional systems.

        The default Coordinate Axes is Inertial, which is ICRF for Earth and Sun.

    axes_epoch: datetime.datetime
        CoordinateAxesEpoch. This is the epoch time for the coordinate axes.

        This parameter is required with coordinate axes that reference an epoch:
            MeanOfEpoch
            TrueOfEpoch
            TEMEOfEpoch
            AlignmentAtEpoch
    
    message: str
        MessageLevel. You can set this value to determine the level of message(s) you receive from STK as it attempts to read the file. The value options are:

            Errors - You only receive the reported error messages.
            Warnings - You receive error messages plus reporting on other information (e.g., unrecognized keywords).
            Verbose - You receive all that is in Warnings plus a success message if STK reads and accepts the complete file.

    body: str
        CentralBody. This is the central body that the attitude points are relative to. The keyword value that completes the phrase can be the name of any registered central body. You can find registered central bodies in the STKData\CentralBodies directory.

        The default is the central body for the vehicle, and that default is Earth.

    int_method: str
        InterpolationMethod. This is the method by which STK interpolates between attitude points. Valid values are:

            Lagrange (Default)
            Hermite - This option enables you to take advantage of angular velocity information to achieve smoother attitude interpolation. One application is if you are using files containing very irregularly spaced data points. Another is if you have data representing a spinning attitude that is susceptible to ringing and aliasing effects if derivative information is not considered during interpolation. Hermitian interpolation assumes that the velocity component of the attitude is the derivative of the position. Hermitian interpolation should not be used when velocity data is not available as part of the attitude file.

    int_order: int
        InterpolationOrder. This is the order of interpolation that STK will use. You can override the default by specifying a different interpolation order. This parameter is often used with the AttitudeLLATimePos attitude format.

        The default is 1.

    Returns
    -------
    a_file: str
        The Attitude File text as a string.
    '''


    format = AttitudeFileFormat(format)
    if message:
        message = MessageLevel(message)

    if axes:
        coord = CoordinateAxes(axes)
        if axes_epoch:
            axes_epoch = CoordinateAxesEpoch(axes_epoch)
        axes = Coordinate(axes=coord, epoch=axes_epoch)

    if epoch:
        epoch = ScenarioEpoch(epoch)

    if body:
        body = CentralBody(body)

    if int_method:
        int_method = InterpolationMethod(int_method)

    if int_order:
        int_order = InterpolationOrder(int_order)

    interpolation = Interpolation(method=int_method, order=int_order)

    if deviations:
        deviations = AttitudeDeviations(deviations)
    
    if blocking:
        blocking = BlockingFactor(blocking)

    a_file = AttitudeFile(
        time,
        data,
        format=format,
        axes=axes,
        epoch=epoch,
        message=message,
        body=body,
        interp=interpolation,
        deviations=deviations,
        blocking=blocking,
    )

    return a_file.to_string()



import numpy as np
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Tuple


class InvalidDataError(Exception):
    pass


class DataValidator(ABC):
    def validate(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError


class NoValidator(DataValidator):
    def validate(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return time, data


@dataclass
class DataShapeValidator(DataValidator):
    ncols: Optional[int] = None
    nrows: Optional[int] = None

    def validate(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        nrows, ncols = data.shape
        if self.ncols is not None and self.ncols != ncols:
            raise InvalidDataError

        if self.nrows is not None and self.nrows != nrows:
            raise InvalidDataError

        return time, data


@dataclass
class QuaternionValidator(DataValidator):
    tolerance: float = 1e-5
    shape_validator: DataShapeValidator = DataShapeValidator(ncols=4)

    def validate(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self.shape_validator.validate(time, data)

        rss = np.abs(np.sqrt(np.sum(data ** 2, axis=1)) - 1)
        print(rss)
        idx = rss < self.tolerance

        if np.sum(idx) == 0:
            raise InvalidDataError

        return time[idx], data[idx, :]


@dataclass
class AngleValidator(DataValidator):
    max_angle: float = 360
    shape_validator: DataShapeValidator = DataShapeValidator(ncols=3)

    def validate(self, time: np.ndarray, data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self.shape_validator.validate(time, data)

        max = np.abs(np.max(data, axis=1))
        idx = max < self.max_angle

        if np.sum(idx) == 0:
            raise InvalidDataError

        return time[idx], data[idx, :]



import datetime
import numpy as np
import pathlib
from typing import Union
from dateutil.parser import parse as parse_datetime, ParserError
from systemstoolkit.typing import DateTimeLike


STK_DATE_FMT = '%d %b %Y %H:%M:%S.%f'

def stk_datetime(timestamp: DateTimeLike) -> str:

    if isinstance(timestamp, datetime.datetime):
        return timestamp.strftime(STK_DATE_FMT)[0:24]

    elif isinstance(timestamp, np.datetime64):
        return timestamp.astype(datetime.datetime).strftime(STK_DATE_FMT)[0:24]

    raise TypeError(
        f'Expected {DateTimeLike}, got "{timestamp}" of type {type(timestamp)}'
    )


def parse_file_data(file_text) -> tuple:
    time, data = [], []
    epoch = None
    for line in file_text.splitlines():
        try:
            parts = line.strip().split()
            
            if parts[0] == 'ScenarioEpoch':
                epoch = parse_datetime(' '.join(parts[1:]))
            
            t = epoch + datetime.timedelta(seconds=float(parts[0]))
            time.append(t)
            data.append([float(x) for x in parts[1:]])
        except (ValueError, ParserError):
            pass
    time = np.array(time, dtype='datetime64[ms]')
    data = np.array(data, dtype='float32')
    return time, data


def read_file_data(file) -> tuple:
    return parse_file_data(pathlib.Path(file).read_text())
